(* --------------------------------------------------------------- *)
(** * Extensions to the Coq library                                *)
(* --------------------------------------------------------------- *)

Set Implicit Arguments.

(* --------------------------------------------------------------- *)
(** ** Axioms                                                      *)
(* --------------------------------------------------------------- *)

Axiom extensionality : forall (X Y : Set) (f g : X -> Y),
  (forall x, f x = g x) -> f = g.

Tactic Notation "extens" := apply extensionality; intro.
Tactic Notation "extens" ident(id) := apply extensionality; intro id.

Axiom proof_irrelevance : forall (A : Prop) (H1 H2 : A), H1 = H2.

(* --------------------------------------------------------------- *)
(** ** Empty                                                       *)
(* --------------------------------------------------------------- *)

Inductive empty : Set := .

Definition empty_inc (X : Set) : empty -> X :=
  fun x => match x with end.

(* --------------------------------------------------------------- *)
(** ** Poor man congruence tactic                                  *)
(* --------------------------------------------------------------- *)

Lemma congr : forall (A B : Type) (f : A -> B) (x y : A),
  x = y -> f x = f y.
Proof.
  destruct 1; trivial.
Qed.

Lemma congr2 : forall (A B C : Type) (f : A -> B -> C)
  (x1 x2 : A) (y1 y2 : B),
  x1 = x2 -> y1 = y2 -> f x1 y1 = f x2 y2.
Proof.
  destruct 1; destruct 1; trivial.
Qed.

Ltac congr :=
  match goal with
  | |- ?f ?a = ?f ?b =>
      change ((fun u => f a = f u) b);
      replace b with a; trivial with *
  | |- ?f ?a ?a2= ?f ?b ?a2  =>
      change ((fun u => f a a2 = f u a2) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a= ?f ?b1 ?b =>
      change ((fun u => f a1 a = f b1 u) b);
      replace b with a; trivial with *
  | |- ?f ?a ?a2 ?a3 = ?f ?b ?a2 ?a3 =>
      change ((fun u => f a a2 a3 = f u a2 a3) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a ?a3 = ?f ?b1 ?b ?a3 =>
      change ((fun u => f a1 a a3 = f b1 u a3) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a = ?f ?b1 ?b2 ?b =>
      change ((fun u => f a1 a2 a = f b1 b2 u) b);
      replace b with a; trivial with *
  | |- ?f ?a ?a2 ?a3 ?a4 = ?f ?b ?a2 ?a3 ?a4 =>
      change ((fun u => f a a2 a3 a4 = f u a2 a3 a4) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a ?a3 ?a4 = ?f ?b1 ?b ?a3 ?a4 =>
      change ((fun u => f a1 a a3 a4 = f b1 u a3 a4) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a ?a4 = ?f ?b1 ?b2 ?b ?a4 =>
      change ((fun u => f a1 a2 a a4 = f b1 b2 u a4) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a3 ?a = ?f ?b1 ?b2 ?b3 ?b =>
      change ((fun u => f a1 a2 a3 a = f b1 b2 b3 u) b);
      replace b with a; trivial with *
  | |- ?f ?a ?a2 ?a3 ?a4 ?a5 = ?f ?b ?a2 ?a3 ?a4 ?a5 =>
      change ((fun u => f a a2 a3 a4 a5 = f u a2 a3 a4 a5) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a ?a3 ?a4 ?a5 = ?f ?a1 ?b ?a3 ?a4 ?a5 =>
      change ((fun u => f a1 a a3 a4 a5 = f a1 u a3 a4 a5) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a ?a4 ?a5 = ?f ?a1 ?a2 ?b ?a4 ?a5 =>
      change ((fun u => f a1 a2 a a4 a5 = f a1 a2 u a4 a5) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a3 ?a ?a5 = ?f ?a1 ?a2 ?a3 ?b ?a5 =>
      change ((fun u => f a1 a2 a3 a a5 = f a1 a2 a3 u a5) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a3 ?a4 ?a = ?f ?a1 ?a2 ?a3 ?a4 ?b =>
      change ((fun u => f a1 a2 a3 a4 a = f a1 a2 a3 a4 u) b);
      replace b with a; trivial with *
  | |- ?f ?a ?a2 ?a3 ?a4 ?a5 ?a6 = ?f ?b ?a2 ?a3 ?a4 ?a5 ?a6 =>
      change ((fun u => f a a2 a3 a4 a5 a6 = f u a2 a3 a4 a5 a6) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a ?a3 ?a4 ?a5 ?a6 = ?f ?a1 ?b ?a3 ?a4 ?a5 ?a6 =>
      change ((fun u => f a1 a a3 a4 a5 a6 = f a1 u a3 a4 a5 a6) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a ?a4 ?a5 ?a6 = ?f ?a1 ?a2 ?b ?a4 ?a5 ?a6 =>
      change ((fun u => f a1 a2 a a4 a5 a6 = f a1 a2 u a4 a5 a6) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a3 ?a ?a5 ?a6 = ?f ?a1 ?a2 ?a3 ?b ?a5 ?a6 =>
      change ((fun u => f a1 a2 a3 a a5 a6 = f a1 a2 a3 u a5 a6) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a3 ?a4 ?a ?a6 = ?f ?a1 ?a2 ?a3 ?a4 ?b ?a6 =>
      change ((fun u => f a1 a2 a3 a4 a a6 = f a1 a2 a3 a4 u a6) b);
      replace b with a; trivial with *
  | |- ?f ?a1 ?a2 ?a3 ?a4 ?a5 ?a = ?f ?a1 ?a2 ?a3 ?a4 ?a5 ?b =>
      change ((fun u => f a1 a2 a3 a4 a5 a = f a1 a2 a3 a4 a5 u) b);
      replace b with a; trivial with *
  end.


Hint Extern 1 (?x = ?y) => congr.

(* --------------------------------------------------------------- *)
(** ** Bijections                                                  *)
(* --------------------------------------------------------------- *)

Record Bij (X Y : Set) : Set := {
  bij_fw :> X -> Y;
  bij_bw : Y -> X;
  bij_l : forall x, bij_bw (bij_fw x) = x;
  bij_r : forall y, bij_fw (bij_bw y) = y
}.

Definition bij_inv (X Y : Set) (f : Bij X Y) : Bij Y X :=
  Build_Bij (bij_bw f) (bij_fw f) (bij_r f) (bij_l f).

Lemma bij_left : forall X Y (f : Bij X Y) x, bij_inv f (f x) = x.
Proof.
apply bij_l.
Qed.

Lemma bij_right : forall X Y (f : Bij X Y) y, f (bij_inv f y) = y.
Proof.
apply bij_r.
Qed.

Lemma bij_inj : forall X Y (f : Bij X Y) x y,
  f x = f y -> x = y.
Proof.
intros.
rewrite <- (bij_l f) with (x := x).
rewrite <- (bij_l f) with (x := y).
rewrite H.
reflexivity.
Qed.

Definition id_bij X : Bij X X :=
  Build_Bij (fun u : X => u) (fun u : X => u)
    (fun u => refl_equal u)
    (fun u => refl_equal u).

Lemma bij_comp : forall X Y (f : Bij X Y) Z (g : Bij Y Z), Bij X Z.
Proof.
intros.
split with (fun x => g (f x)) (fun y => bij_inv f (bij_inv g y)).
abstract (intros; do 2 rewrite bij_left; reflexivity).
abstract (intros; do 2 rewrite bij_right; reflexivity).
Defined.

(* --------------------------------------------------------------- *)
(** ** More terms about [option]                                   *)
(* --------------------------------------------------------------- *)

Definition default (X Y : Set) (f : X -> Y)
    (y : Y) (x : option X) : Y :=
  match x with
  | None => y
  | Some u => f u
  end.

Definition optfct (X Y : Set) (f : X -> Y)
    (x : option X) : option Y :=
  match x with
  | None => None
  | Some u => Some (f u)
  end.

Definition optfct_bij X Y (f : Bij X Y) : Bij (option X) (option Y).
Proof.
intros.
split with (optfct f) (optfct (bij_inv f)).
abstract (destruct x; simpl; try rewrite bij_l; reflexivity).
abstract (destruct y; simpl; try rewrite bij_r; reflexivity).
Defined.

(* --------------------------------------------------------------- *)
(** ** More terms about [sum]                                      *)
(* --------------------------------------------------------------- *)

Definition sumfct
    (X1 Y1 : Set) (f1 : X1 -> Y1) (X2 Y2 : Set) (f2 : X2 -> Y2)
    (u : X1 + X2) : Y1 + Y2 :=
  match u with inl x => inl Y2 (f1 x) | inr y => inr Y1 (f2 y) end.

Definition unempty_l (X : Set) (u : empty + X) : X :=
  match u with inl x => empty_inc _ x | inr x => x  end.

Definition unempty_r (X : Set) (x : X + empty) : X :=
  match x with inl x => x | inr x => empty_inc _ x end.

Definition to_empty_l (X : Set) (x : X) : empty + X :=
  inr _ x.

Definition to_empty_r (X : Set) (x : X) : X + empty :=
  inl _ x.

Definition sumfct_l (X1 X2 Y : Set) (f : X1 -> X2) : X1 + Y -> X2 + Y :=
  fun u =>
  match u with
  | inl x => inl _ (f x)
  | inr y => inr _ y
  end.

Definition sumfct_r (X Y1 Y2 : Set) (f : Y1 -> Y2) : X + Y1 -> X + Y2 :=
  fun u =>
  match u with
  | inl x => inl _ x
  | inr y => inr _ (f y)
  end.

Definition sum_assoc_r X Y Z : (X + Y) + Z -> X + (Y + Z) :=
  fun u =>
    match u with
    | inl (inl x) => inl _ x
    | inl (inr y) => inr _ (inl _ y)
    | inr z => inr _ (inr _ z)
    end.

Definition sum_assoc_l X Y Z : X + (Y + Z) -> (X + Y) + Z :=
  fun u =>
    match u with
    | inl x => inl _ (inl _ x)
    | inr (inl y) => inl _ (inr _ y)
    | inr (inr z) => inr _ z
    end.

Definition sum_assoc X Y Z : Bij ((X + Y) + Z) (X + (Y + Z)).
Proof.
intros.
split with (@sum_assoc_r X Y Z) (@sum_assoc_l X Y Z).
abstract (destruct x as [[x|y]|z]; reflexivity).
abstract (destruct y as [x|[y|z]]; reflexivity).
Defined.

Definition opt_right X Y : option (X + Y) -> X + option Y :=
  default
    (fun u =>
      match u with inl x => inl _ x | inr y => inr _ (Some y) end)
    (inr _ None).

Lemma opt_right_inj : forall X Y (x y : option (X + Y)),
  opt_right x = opt_right y -> x = y.
Proof.
destruct x as [[u|v]|]; destruct y as [[p|q]|]; simpl;
  try discriminate 1; inversion 1; subst; reflexivity.
Qed.

Definition opt_unright X Y : X + option Y -> option (X + Y) :=
  fun u =>
    match u with
    | inl x => Some (inl _ x)
    | inr (Some y) => Some (inr _ y)
    | inr None => None
    end.

Definition opt_right_bij X Y : Bij (option (X + Y)) (X + option Y).
Proof.
intros.
split with (@opt_right X Y) (@opt_unright X Y).
abstract (destruct x as [[x|y]|]; reflexivity).
abstract (destruct y as [x|[y|]]; reflexivity).
Defined.

(* --------------------------------------------------------------- *)
(** ** Fibered product                                             *)
(* --------------------------------------------------------------- *)

(* --------------------------------------------------------------- *)
(** We define the fibered product for [Set] and prove some
    basic properties.                                              *)
(* --------------------------------------------------------------- *)

Section FProd.

Variable (X1 X2 B U : Set).
Variable (f1 : X1 -> B) (f2 : X2 -> B).
Variable (p1 : U -> X1) (p2 : U -> X2).

Record FProd : Type :=
{
  fpair : forall (x1 : X1) (x2 : X2), f1 x1 = f2 x2 -> U;
  fprod_comm : forall u, f1 (p1 u) = f2 (p2 u);
  fpair_1 : forall x1 x2 (H : f1 x1 = f2 x2),
     p1 (fpair x1 x2 H) = x1;
  fpair_2 : forall x1 x2 (H : f1 x1 = f2 x2),
     p2 (fpair x1 x2 H) = x2;
  fprod_unique : forall u v, p1 u = p1 v -> p2 u = p2 v -> u = v
}.

Variable FP : FProd.

Lemma fpair_unique : forall u x1 x2 (H : f1 x1 = f2 x2),
  p1 u = x1 -> p2 u = x2 -> u = fpair FP (* x1 x2 *) H.
Proof.
intros. apply (fprod_unique FP).
rewrite fpair_1. assumption.
rewrite fpair_2. assumption.
Qed.

Definition fprod_map (W : Set) (q1 : W -> X1) (q2 : W -> X2)
    (H : forall w, f1 (q1 w) = f2 (q2 w)) (w : W) : U :=
  fpair FP (* (q1 w) (q2 w) *) (H w).

Lemma fprod_map_1 : forall (W : Set) (q1 : W -> X1) (q2 : W -> X2)
    (H : forall w, f1 (q1 w) = f2 (q2 w)) (w : W),
  p1 (fprod_map q1 q2 H w) = q1 w.
Proof.
intros. unfold fprod_map. rewrite fpair_1. reflexivity.
Qed.

Lemma fprod_map_2 : forall (W : Set) (q1 : W -> X1) (q2 : W -> X2)
    (H : forall w, f1 (q1 w) = f2 (q2 w)) (w : W),
  p2 (fprod_map q1 q2 H w) = q2 w.
Proof.
intros. unfold fprod_map. rewrite fpair_2. reflexivity.
Qed.

Lemma fprod_map_unique : forall (W : Set) (q1 : W -> X1) (q2 : W -> X2)
    (H : forall w, f1 (q1 w) = f2 (q2 w)) (h : W -> U),
  (forall w, p1 (h w) = q1 w) -> (forall w, p2 (h w) = q2 w) ->
  h = (fprod_map q1 q2 H).
Proof.
intros. extens w. unfold fprod_map.
apply (fprod_unique FP).
rewrite fpair_1. trivial.
rewrite fpair_2. trivial.
Qed.

End FProd.

Section Mk_FProd.

Variable (X1 X2 B : Set).
Variable (f1 : X1 -> B) (f2 : X2 -> B).

Definition mkfprod_set : Set :=
  { x : X1 * X2 | f1 (fst x) = f2 (snd x) }.

Definition mkfprod_p1 (x : mkfprod_set) : X1 :=
  let (x, _) := x in (fst x).

Definition mkfprod_p2 (x : mkfprod_set) : X2 :=
  let (x, _) := x in (snd x).

Definition mkfprod : FProd f1 f2 mkfprod_p1 mkfprod_p2.
Proof.
split with (fun x1 x2 (H : f1 x1 = f2 x2) =>
              exist (fun x => f1 (fst x) = f2 (snd x)) (x1, x2) H);
  simpl; try reflexivity.
destruct u as [[u1 u2] Hu]. simpl in *. assumption.
destruct u as [[u1 u2] Hu]. destruct v as [[v1 v2] Hv]. simpl.
destruct 1. destruct 1. congr. apply proof_irrelevance.
Qed.

End Mk_FProd.

Section FProd_option.

Variable (X1 X2 B U : Set).
Variable (f1 : X1 -> B) (f2 : X2 -> B).
Variable (p1 : U -> X1) (p2 : U -> X2).
Variable FP : FProd f1 f2 p1 p2.

Let sig_h x1 x2 (H : optfct f1 x1 = optfct f2 x2) :
  { u : option U | optfct p1 u = x1 /\ optfct p2 u = x2 }.
Proof.
destruct x1; destruct x2; simpl; try discriminate 1; intros.
assert (He : f1 x = f2 x0).
inversion H; reflexivity.
split with (Some (fpair FP x x0 He)).
simpl. rewrite fpair_1. rewrite fpair_2. auto.
split with (@None U). simpl. auto.
Qed.

Let h x1 x2 (H : optfct f1 x1 = optfct f2 x2) : option U :=
  let (u, _) := sig_h x1 x2 H in u.

Definition FProd_option :
  FProd (optfct f1) (optfct f2) (optfct p1) (optfct p2).
Proof.
split with h; intros; unfold h.
destruct u; simpl. rewrite (fprod_comm FP). reflexivity. reflexivity.
destruct (sig_h x1 x2 H). tauto.
destruct (sig_h x1 x2 H). tauto.
destruct u; destruct v; simpl in H, H0; try discriminate H.
congr.
apply (fprod_unique FP).
inversion H. reflexivity.
inversion H0. reflexivity.
reflexivity.
Qed.

End FProd_option.

Section Sum_Dec.

Variable X Y : Set.

Inductive in_left : X + Y -> Prop :=
  | in_left_inl : forall x, in_left (inl Y x).

Inductive in_right : X + Y -> Prop :=
  | in_right_inr : forall y, in_right (inr X y).

Lemma not_in_left_inr : forall (y : Y),
  ~ in_left (inr X y).
Proof.
red. inversion 1.
Qed.

Lemma not_in_right_inl : forall (x : X),
  ~ in_right (inl Y x).
Proof.
red. inversion 1.
Qed.

Hint Resolve in_left_inl in_right_inr not_in_left_inr not_in_right_inl.

Definition send_left (z : X + Y) : in_left z -> X :=
  match z as s return (in_left s -> X) with
  | inl x => fun _ : in_left (inl Y x) => x
  | inr y =>
     fun H : in_left (inr X y) => False_rec X (not_in_left_inr H)
  end.

Definition send_right (z : X + Y) : in_right z -> Y :=
  match z as s return (in_right s -> Y) with
  | inr y => fun _ : in_right (inr X y) => y
  | inl x =>
     fun H : in_right (inl Y x) => False_rec Y (not_in_right_inl H)
  end.

Lemma send_left_eq : forall (x : X) (z : X + Y) (Hz : in_left z),
  inl Y x = z -> send_left Hz = x.
Proof.
destruct z; simpl.
inversion 2.
reflexivity.
discriminate 1.
Qed.

Lemma send_right_eq : forall (y : Y) (z : X + Y) (Hz : in_right z),
  inr X y = z -> send_right Hz = y.
Proof.
destruct z; simpl.
discriminate 1.
inversion 2.
reflexivity.
Qed.

Lemma inl_send_left : forall (z : X + Y) (Hz : in_left z),
  inl Y (send_left Hz) = z.
Proof.
destruct z. reflexivity. simpl. inversion Hz.
Qed.

Lemma inr_send_right : forall (z : X + Y) (Hz : in_right z),
  inr X (send_right Hz) = z.
Proof.
destruct z. simpl. inversion Hz. reflexivity.
Qed.

Lemma not_in_left_imp_in_right : forall z : X + Y,
  ~ in_left z -> in_right z.
Proof.
destruct z; firstorder.
elimtype False.
firstorder.
Qed.

Definition in_left_dec : forall z : X + Y,
  { in_left z } + { ~ in_left z }.
Proof.
destruct z.
left. constructor.
right. abstract (intro; inversion H).
Defined.

Definition in_right_dec : forall z : X + Y,
  { in_right z } + { ~ in_right z }.
Proof.
destruct z.
right. abstract (intro; inversion H).
left. constructor.
Defined.

Definition in_left_right : forall z : X + Y,
  { in_left z } + { in_right z }.
Proof.
destruct z.
left. constructor.
right. constructor.
Defined.

End Sum_Dec.

Section Sum_Bij.

Variable (Z : Set) (P : Z -> Prop).
Variable P_dec : forall z, { P z } + { ~P z }.

Definition Sum_Bij_1 : Set := { z : Z | P z }.
Definition Sum_Bij_2 : Set := { z : Z | ~ P z }.

Definition sum_bij_1 (z : Z) (H : P z) : Sum_Bij_1 :=
  exist (fun z => P z) z H.

Definition sum_bij_2 (z : Z) (H : ~ P z) : Sum_Bij_2 :=
  exist (fun z => ~ P z) z H.

Definition sum_bij_pair (u : Sum_Bij_1 + Sum_Bij_2) : Z :=
  match u with
  | inl (exist x _) => x
  | inr (exist y _) => y
  end.

Definition sum_bij_dest (z : Z) : Sum_Bij_1 + Sum_Bij_2 :=
  match P_dec z with
  | left p => inl Sum_Bij_2 (sum_bij_1 p)
  | right p => inr Sum_Bij_1 (sum_bij_2 p)
  end.

Lemma sum_bij_left : forall x : Sum_Bij_1 + Sum_Bij_2,
  sum_bij_dest (sum_bij_pair x) = x.
Proof.
destruct x as [[x Hx] | [y Hy]]; simpl.
unfold sum_bij_dest.
destruct (P_dec x).
replace p with Hx. reflexivity. apply proof_irrelevance.
elimtype False. tauto.

unfold sum_bij_dest.
destruct (P_dec y).
elimtype False. tauto.
replace n with Hy. reflexivity. apply proof_irrelevance.
Qed.

Lemma sum_bij_right : forall y : Z,
  sum_bij_pair (sum_bij_dest y) = y.
Proof.
intros. unfold sum_bij_dest.
destruct (P_dec y); reflexivity.
Qed.

Definition sum_bij : Bij (Sum_Bij_1 + Sum_Bij_2) Z.
Proof.
split with sum_bij_pair sum_bij_dest.
apply sum_bij_left.
apply sum_bij_right.
Defined.

Lemma sum_bij_prop_1 : forall x : Sum_Bij_1,
  P (sum_bij (inl _ x)).
Proof.
destruct x as [z H]. assumption.
Qed.

Lemma sum_bij_prop_2 : forall y : Sum_Bij_2,
  ~ P (sum_bij (inr _ y)).
Proof.
destruct y as [z H]. assumption.
Qed.

Lemma sum_bij_inl : forall (x : Sum_Bij_1),
  sum_bij (inl _ x) = proj1_sig x.
Proof.
destruct x. reflexivity.
Qed.

Lemma sum_bij_inr : forall (y : Sum_Bij_2),
  sum_bij (inr _ y) = proj1_sig y.
Proof.
destruct y. reflexivity.
Qed.

Lemma proj1_sig_send_left : forall (z : Sum_Bij_1 + Sum_Bij_2)
    (Hz : in_left z),
  proj1_sig (send_left Hz) = sum_bij z.
Proof.
destruct z as [[x Hx] | [y Hy]]; simpl. reflexivity. inversion Hz.
Qed.

Lemma proj1_sig_send_right : forall (z : Sum_Bij_1 + Sum_Bij_2)
    (Hz : in_right z),
  proj1_sig (send_right Hz) = sum_bij z.
Proof.
destruct z as [[x Hx] | [y Hy]]; simpl. inversion Hz. reflexivity.
Qed.

End Sum_Bij.
