(* --------------------------------------------------------------- *)
(** * Well formed environments                                     *)
(* --------------------------------------------------------------- *)

Set Implicit Arguments.

Require Export F_sub_facts.

(* --------------------------------------------------------------- *)
(** The following relations says when an environment [d:env X1 Y1]
    is  well formed with respect to an environment [e:env X1 Y1].  *)
(* --------------------------------------------------------------- *)


Inductive WF : forall X1 Y1 X2 Y2 : Set,
    (X1 -> X2) -> (Y1 -> Y2) -> env X1 Y1 -> env X2 Y2 -> Prop :=
  | WF_intro : forall (X1 X2 Y1 Y2 : Set)
      (e : env X1 Y1) (d : env X2 Y2)
      (r : X1 -> X2) (f : Y1 -> Y2) (g : Y2 -> T (X1 + Y1)),
    (forall x, g (f x) = e x) ->
    (forall y, fct (sumfct r f) (g y) = d y) -> WF r f e d
  | WF_trs : forall X1 Y1 X2 Y2 X3 Y3
       (e : env X1 Y1) (c : env X2 Y2) (d : env X3 Y3)
       (r' : X1 -> X2) (r'' : X2 -> X3) (r : X1 -> X3)
       (f : Y1 -> Y2) (g : Y2 -> Y3) (h : Y1 -> Y3),
     (forall u, h u = g (f u)) ->
     (forall u, r u = r'' (r' u)) ->
     WF r' f e c -> WF r'' g c d -> WF r h e d.
Hint Constructors WF.

(* --------------------------------------------------------------- *)
(* The following relation [WF0] is equivalent to WF in the special
   case where X1 and X2 are empty.                                 *)
(* --------------------------------------------------------------- *)

Inductive WF0 : forall X Y : Set,
      (X -> Y) -> env0 X -> env0 Y -> Prop :=
  | WF0_intro : forall (X Y : Set) (e : env0 X) (d : env0 Y)
      (f : X -> Y) (g : Y -> T X),
    (forall x, g (f x) = e x) ->
    (forall y, fct f (g y) = d y) -> WF0 f e d
  | WF0_trs : forall (X Y Z : Set)
       (e : env0 X) (c : env0 Y) (d : env0 Z)
       (f : X -> Y) (g : Y -> Z) (h : X -> Z),
     (forall u, h u = g (f u)) ->
     WF0 f e c -> WF0 g c d -> WF0 h e d.
Hint Constructors WF0.

(* --------------------------------------------------------------- *)
(** This is the absolute definition of well formed.                *)
(* --------------------------------------------------------------- *)

Definition WF0all (X : Set) (g : env0 X) : Prop :=
  WF0 (empty_inc _) (empty_inc _) g.

(* --------------------------------------------------------------- *)
(** ** Equivalence of WF with [X1 = X2] and WF0                    *)
(* --------------------------------------------------------------- *)

(* --------------------------------------------------------------- *)
(** The easy part: [WF0] implies [WF]                              *)
(* --------------------------------------------------------------- *)

Lemma WF0_WF : forall (X Y : Set) (f : X -> Y)
    (e : env0 X) (d : env0 Y),
  WF0 f e d -> WF (empty_inc empty) f (env_of_env0 e) (env_of_env0 d).
Proof.
induction 1.
apply WF_intro with (g := fun u : Y => fct (@to_empty_l _) (g u)).
intros. rewrite H. reflexivity.
intros. rewrite fct_fct. unfold env_of_env0.
rewrite <- H0. rewrite fct_fct. reflexivity.
apply WF_trs with
  (c := env_of_env0 c) (f := f) (g := g)
  (r' := empty_inc empty) (r'' := empty_inc empty).
assumption. tauto. assumption. assumption.
Qed.

(* --------------------------------------------------------------- *)
(** The had part: [WF] implies [WF0].                              *)
(* --------------------------------------------------------------- *)

Section WF_WF0.

(* --------------------------------------------------------------- *)
(** Cross fertilization for WF_WF0                                 *)
(* --------------------------------------------------------------- *)

Let WF_WF0_xfer : forall (U X V Y : Set)
    (r : U -> V) (f : X -> Y) (e : env U X) (d : env V Y),
  WF r f e d ->
  forall (p : U -> empty) (q : V -> empty)
    (e' : env0 X) (d' : env0 Y),
  (forall x : X, (fct (fun u => inr U u) (e' x)) = e x) ->
  (forall y : Y, (fct (fun u => inr V u) (d' y)) = d y) ->
  WF0 f e' d'.
Proof.
induction 1.
intros.
apply WF0_intro with
  (g := fun y : Y2 =>
          fct (@unempty_l Y1) (fct (sumfct p (fun y : Y1 => y)) (g y))).
intros. rewrite fct_fct. rewrite H. rewrite <- H1.
rewrite fct_fct. apply fct_id. reflexivity.
intros. do 2 rewrite fct_fct.
transitivity
  (fct (@unempty_l Y2)
    (fct (sumfct q (fun y : Y2 => y)) (fct (sumfct r f) (g y)))).
do 2 rewrite fct_fct. congr. extens x. destruct x.
destruct (p x). reflexivity.
rewrite H0. rewrite fct_fct. rewrite <- H2.
rewrite fct_fct. apply fct_id. reflexivity.
intros.
apply WF0_trs with  (f := f) (g := g)
  (c := fun y : Y2 =>
          fct (@unempty_l _)
            (fct (sumfct (fun x => q (r'' x)) (fun y : Y2 => y)) (c y))).
assumption.
apply IHWF1.
assumption. auto. assumption.
intros. do 2 rewrite fct_fct.
apply fct_id. destruct u. destruct (q (r'' x)). reflexivity.
apply IHWF2. auto. assumption.
intros. do 2 rewrite fct_fct.
apply fct_id. destruct u. destruct (q (r'' x0)).
reflexivity. assumption.
Qed.

Lemma WF_WF0 : forall (X Y : Set) (f : X -> Y)
    (e : env0 X) (d : env0 Y),
  WF (empty_inc _) f (env_of_env0 e) (env_of_env0 d) ->
  WF0 f e d.
Proof.
intros.
apply WF_WF0_xfer with (r := empty_inc empty)
  (e := env_of_env0 e) (d := env_of_env0 d);
  trivial.
Qed.

End WF_WF0.

Let WF0_commute : forall (X Y : Set) (f : X -> Y) g d,
  WF0 f g d ->
  forall x, fct f (g x) = d (f x).
Proof.
induction 1.
intros. rewrite <- H. rewrite H0. reflexivity.
intros.
transitivity (fct g (fct f (e x))).
rewrite fct_fct. congr. extens. trivial.
rewrite IHWF0_1. rewrite IHWF0_2. rewrite H. reflexivity.
Qed.

(* --------------------------------------------------------------- *)
(** An environment which extends to a well-defined environment
    is itself well-defined.                                        *)
(* --------------------------------------------------------------- *)

Section WF0all_Sub.

(* --------------------------------------------------------------- *)
(** Cross fertilization for WF0all_sub.                            *)
(* --------------------------------------------------------------- *)

Let WF0_sub_xfer : forall (Y Z : Set)
    (r : Z -> Y) (d : env0 Y) (e : env0 Z) (Hr : WF0 r e d),
  forall (X W : Set) (f : X -> Y) (p : W -> X) (q : W -> Z)
    (g : env0 X) (b : env0 W)
    (Hf : forall x : X, fct f (g x) = d (f x))
    (Hp : forall w : W, fct p (b w) = g (p w))
    (Hq : forall w : W, fct q (b w) = e (q w))
    (FP : FProd f r p q),
  WF0 p b g.
Proof.
induction 1; intros.
rename X0 into Z. rename f0 into r.
assert (FPT := FProd_T FP).
assert (Hcomm : forall z : Z, fct r (g0 z) = fct f (g (r z))).
intros. rewrite H0. rewrite Hf. reflexivity.
apply WF0_intro with
  (g := fun z : Z =>
          fpair FPT (g0 z) (g (r z)) (Hcomm z)).
intros. apply (fprod_unique FPT).
rewrite fpair_1. rewrite Hp. reflexivity.
rewrite fpair_2. rewrite (fprod_comm FP).
rewrite H. rewrite Hq. reflexivity.
intros. rewrite fpair_1. reflexivity.
rename Y into U. rename Z into Y. rename X0 into Z.
pose (V := mkfprod_set f0 g).
pose (p0 := mkfprod_p1 (f1 := f0) (f2 := g)).
pose (q0 := mkfprod_p2 (f1 := f0) (f2 := g)).
pose (FP0 := mkfprod f0 g).
assert (Hcomm : forall v, fct f0 (g0 (p0 v)) = fct g (c (q0 v))).
intros. rewrite Hf.
unfold p0. rewrite (fprod_comm FP0).
rewrite (WF0_commute Hr2). reflexivity.
pose (b0 := fun v : V =>
              fpair (FProd_T FP0) (g0 (p0 v)) (c (q0 v)) (Hcomm v)).
pose (q1 := fun w => f (q w)).
assert (Hcomm2 : forall w : W, f0 (p w) = g (q1 w)).
intros. rewrite (fprod_comm FP). rewrite H. reflexivity.
pose (p1 := fun w : W => fpair FP0 (p w) (q1 w) (Hcomm2 w)).
apply WF0_trs with (c := b0) (g := p0) (f := p1).
intros. unfold p0, p1. rewrite fpair_1. reflexivity.
apply IHHr1 with (f0 := q0) (q := q).
intros. unfold q0, b0. rewrite fpair_2. reflexivity.
intros. unfold b0, p1.
apply (fpair_unique (FProd_T FP0)).
rewrite fct_fct. unfold p0, q1. rewrite fpair_1.
rewrite <- Hp. congr. extens. rewrite fpair_1. reflexivity.
rewrite fct_fct. unfold q0, q1. rewrite fpair_2.
transitivity (fct f (fct q (b w))).
rewrite fct_fct. congr. extens. rewrite fpair_2. reflexivity.
rewrite Hq. rewrite (WF0_commute Hr1). reflexivity.
assumption.
assert (Hcomm3 : forall v x, q0 v = f x -> f0 (p0 v) = h x).
unfold p0, q0. intros.
rewrite (fprod_comm (mkfprod f0 g)).
rewrite H0. rewrite H. reflexivity.
split with
  (fun (v : V) (x : X) (H : q0 v = f x) =>
    fpair FP (p0 v) x (Hcomm3 v x H)).
intros. unfold q0, p1. rewrite fpair_2. reflexivity.
intros. unfold p1. apply (fprod_unique FP0).
rewrite fpair_1. rewrite fpair_1. reflexivity.
rewrite fpair_2. unfold q1. rewrite fpair_2.
rewrite <- H0. reflexivity.
intros. rewrite fpair_2. reflexivity.
unfold p1. intros. apply (fprod_unique FP).
transitivity (mkfprod_p1 (f1 := f0) (f2 := g)
  (fpair FP0 (p u) (q1 u) (Hcomm2 u))).
rewrite fpair_1. reflexivity.
rewrite H0. rewrite fpair_1. reflexivity.
assumption.
apply IHHr2 with (f := f0) (q := q0).
assumption.
intros. unfold p0, b0, p0. rewrite fpair_1. reflexivity.
intros. unfold q0, b0. rewrite fpair_2. reflexivity.
apply FP0.
Qed.

Lemma WF0all_sub : forall (X Y : Set) (f : X -> Y)
    (d : env0 Y) (g : env0 X),
  (forall x : X, fct f (g x) = d (f x)) ->
  WF0all d -> WF0all g.
Proof.
intros. unfold WF0all.
apply WF0_sub_xfer with (f := f) (r := empty_inc Y)
  (p := empty_inc X) (q := empty_inc empty)
  (d := d) (e := empty_inc (T empty)).
assumption. assumption. destruct w. destruct w.
split with
  (fun (x : X) (a : empty) (H : f x = empty_inc Y a) => a).
destruct u. destruct x2. destruct x2. destruct u.
Qed.

End WF0all_Sub.

(* --------------------------------------------------------------- *)
(** Consenv respects well-formedness                               *)
(* --------------------------------------------------------------- *)

Lemma WF0_consenv : forall (X : Set) (e : env0 X) (t : T X),
  WF0 (@Some X) e (consenv e t).
Proof.
intros.
apply WF0_intro with (fun x => default e t x).
reflexivity.
unfold consenv, shift. intros.
destruct y; reflexivity.
Qed.

Lemma WF0all_consenv : forall (X : Set) (g : env0 X) (p : T X),
  WF0all g -> WF0all (consenv g p).
Proof.
unfold WF0all. intros.
apply WF0_trs with (f := empty_inc X) (g := (@Some X)) (c := g).
destruct u. assumption. apply WF0_consenv.
Qed.

Lemma consenv_WF0all : forall (X : Set) (g : env0 X) (q : T X),
  WF0all (consenv g q) -> WF0all g.
Proof.
intros.
apply WF0all_sub with (d := consenv g q) (f := @Some X).
reflexivity.
assumption.
Qed.

(* --------------------------------------------------------------- *)
(** Appendenv respects well-formedness                             *)
(* --------------------------------------------------------------- *)

Lemma appendenv_WF0all_1 : forall (X Y : Set)
    (g : env0 X) (d : env X Y),
  WF0all (appendenv g d) -> WF0all g.
Proof.
intros.
apply WF0all_sub with (d := appendenv g d) (f := @inl X Y).
reflexivity.
assumption.
Qed.

(* --------------------------------------------------------------- *)
(** ** Proof of appendenv_WF0all_2                                 *)
(* --------------------------------------------------------------- *)

Section Appendenv_WF0.

Section Appendenv_WF_2_xfer_trs.

Variable (Z' X' Y' Z X Y Z0 : Set).
Variable (b' : Bij (X' + Y') Z') (b : Bij (X + Y) Z).
Variable (g' : env0 Z') (g0 : env0 Z0) (g : env0 Z).
Variable (e' : env X' Y') (e : env X Y).
Variable (f1 : Z' -> Z0) (f2 : Z0 -> Z) (f : Z' -> Z).
Variable (fx : X' -> X) (fy : Y' -> Y).
Hypothesis Hf12 : forall z, f z = f2 (f1 z).
Hypothesis Hf : forall z, f z = b (sumfct fx fy (bij_inv b' z)).

Hypothesis He :
  e = (fun y : Y => fct (bij_inv b) (g (b (inr X y)))).

Hypothesis He' :
  e' = (fun y : Y' => fct (bij_inv b') (g' (b' (inr X' y)))).

Definition P (z : Z0) : Prop := in_left (bij_inv b (f2 z)).

Let P_dec : forall z, { P z } + { ~ P z }.
Proof.
intros. apply (in_left_dec (bij_inv b (f2 z))).
Defined.

Let X0 : Set := Sum_Bij_1 P.
Let Y0 : Set := Sum_Bij_2 P.
Let b0 := sum_bij P P_dec.
Definition f_fx2 (x : X0) : X + Y := bij_inv b (f2 (b0 (inl Y0 x))).

Let H_fx2 : forall x : X0, in_left (f_fx2 x).
Proof.
intros. unfold f_fx2.
change (P (b0 (inl Y0 x))).
unfold b0.
apply (sum_bij_prop_1 P_dec).
Qed.

Let fx2 (x : X0) := send_left (H_fx2 x).

Let f_fy2 (y : Y0) := bij_inv b (f2 (b0 (inr X0 y))).

Let H_fy2 : forall y : Y0, in_right (f_fy2 y).
Proof.
intros. unfold f_fy2.
apply not_in_left_imp_in_right.
change (~ P (b0 (inr X0 y))).
unfold b0.
apply (sum_bij_prop_2 P_dec).
Qed.

Let fy2 (y : Y0) := send_right (H_fy2 y).

Let f_fx1 (x : X') := bij_inv b0 (f1 (b' (inl Y' x))).

Let H_fx1 : forall x : X', in_left (f_fx1 x).
Proof.
unfold f_fx1, b0. simpl. unfold sum_bij_dest. intros.
destruct (P_dec (f1 (b' (inl Y' x)))).
apply in_left_inl. elimtype False.
apply n. unfold P.
rewrite <- Hf12. rewrite Hf.
do 2 rewrite bij_left. simpl.
apply in_left_inl.
Qed.

Let fx1 (x : X') := send_left (H_fx1 x).

Let f_fy1 (y : Y') := bij_inv b0 (f1 (b' (inr X' y))).

Let H_fy1 : forall y : Y', in_right (f_fy1 y).
Proof.
unfold f_fy1, b0. simpl. unfold sum_bij_dest. intros.
destruct (P_dec (f1 (b' (inr X' y)))).
2: apply in_right_inr.
elimtype False.
unfold P in p.
rewrite <- Hf12 in p. rewrite Hf in p.
do 2 rewrite bij_left in p. simpl in p.
inversion p.
Qed.

Let fy1 := fun y : Y' => send_right (H_fy1 y).

Let c (y : Y0) := fct (bij_inv b0) (g0 (b0 (inr X0 y))).

Let fx_comp : forall x : X', fx x = fx2 (fx1 x).
Proof.
unfold fx1, fx2. intros.
symmetry. apply send_left_eq.
unfold f_fx2. rewrite inl_send_left.
unfold f_fx1. rewrite bij_right.
rewrite <- Hf12. rewrite Hf.
do 2 rewrite bij_left. reflexivity.
Qed.

Let fy_comp : forall y : Y', fy y = fy2 (fy1 y).
Proof.
unfold fy1, fy2. intros.
symmetry. apply send_right_eq.
unfold f_fy2. rewrite inr_send_right.
unfold f_fy1. rewrite bij_right.
rewrite <- Hf12. rewrite Hf.
do 2 rewrite bij_left. reflexivity.
Qed.

Let b0_inl_fx1 : forall x : X',
  b0 (inl Y0 (fx1 x)) = f1 (b' (inl Y' x)).
Proof.
intros. unfold b0, fx1.
rewrite sum_bij_inl with (P_dec := P_dec).
rewrite proj1_sig_send_left with (P_dec := P_dec).
unfold f_fx1, b0. rewrite bij_right. reflexivity.
Qed.

Let b0_inr_fy1 : forall y : Y',
  b0 (inr X0 (fy1 y)) = f1 (b' (inr X' y)).
Proof.
intros. unfold b0, fy1.
rewrite sum_bij_inr with (P_dec := P_dec).
rewrite proj1_sig_send_right with (P_dec := P_dec).
unfold f_fy1, b0. rewrite bij_right. reflexivity.
Qed.

Let f1_sum : forall z : Z',
  f1 z = b0 (sumfct fx1 fy1 (bij_inv b' z)).
Proof.
intros.
pose (u := bij_inv b' z).
replace z with (b' u).
2: unfold u; rewrite bij_right; reflexivity.
rewrite bij_left.
destruct u.
change (f1 (b' (inl Y' x)) = b0 (inl _ (fx1 x))).
rewrite b0_inl_fx1.
reflexivity.
change (f1 (b' (inr X' y)) = b0 (inr _ (fy1 y))).
rewrite b0_inr_fy1.
reflexivity.
Qed.

Let f2_sum : forall z : Z0,
  f2 z = b (sumfct fx2 fy2 (bij_inv b0 z)).
Proof.
intros. simpl. unfold sum_bij_dest.
destruct (P_dec z); simpl.
unfold sum_bij_1, fx2. rewrite inl_send_left.
unfold f_fx2. rewrite bij_right. reflexivity.
unfold sum_bij_2, fy2. rewrite inr_send_right.
unfold f_fy2. rewrite bij_right. reflexivity.
Qed.

Hypothesis IHWF0_1 :
  forall (X Y : Set) (b : Bij (X + Y) Z0) (X' Y' : Set)
            (b' : Bij (X' + Y') Z') (fx : X' -> X) (fy : Y' -> Y)
            (e : env X Y) (e' : env X' Y'),
          (forall z : Z', f1 z = b (sumfct fx fy (bij_inv b' z))) ->
          e = (fun y : Y => fct (bij_inv b) (g0 (b (inr X y)))) ->
          e' = (fun y : Y' => fct (bij_inv b') (g' (b' (inr X' y)))) ->
          WF fx fy e' e.

Hypothesis IHWF0_2 :
  forall (X Y : Set) (b : Bij (X + Y) Z) (X' Y' : Set)
            (b' : Bij (X' + Y') Z0) (fx : X' -> X) (fy : Y' -> Y)
            (e : env X Y) (e' : env X' Y'),
          (forall z : Z0, f2 z = b (sumfct fx fy (bij_inv b' z))) ->
          e = (fun y : Y => fct (bij_inv b) (g (b (inr X y)))) ->
          e' = (fun y : Y' => fct (bij_inv b') (g0 (b' (inr X' y)))) ->
          WF fx fy e' e.

Lemma appendenv_WF_2_xfer_trs : WF fx fy e' e.
Proof.
apply WF_trs with (c := c)
  (r' := fx1) (r'' := fx2) (f := fy1) (g := fy2).
apply fy_comp.
apply fx_comp.
apply IHWF0_1 with (b := b0) (b' := b').
apply f1_sum.
reflexivity.
apply He'.
apply IHWF0_2 with (b := b) (b' := b0).
apply f2_sum.
apply He.
reflexivity.
Qed.

End Appendenv_WF_2_xfer_trs.

Let appendenv_WF_2_xfer : forall (Z Z' : Set) (f : Z' -> Z)
    (g : env0 Z) (g' : env0 Z'), 
  WF0 f g' g ->
    forall (X Y : Set) (b : Bij (X + Y) Z)
      (X' Y' : Set) (b' : Bij (X' + Y') Z')
      (fx : X' -> X) (fy : Y' -> Y)
      (e : env X Y) (e' : env X' Y'),
   (forall z, f z = b (sumfct fx fy (bij_inv b' z))) ->
   (e = fun y => fct (bij_inv b) (g (b (@inr X Y y)))) ->
   (e' = fun y => fct (bij_inv b') (g' (b' (@inr X' Y' y)))) ->
   WF fx fy e' e.
Proof.
induction 1.
intros.
apply WF_intro with
  (g := fun y => (fct (bij_inv b') (g (b (@inr X0 Y0 y))))).
intros. subst e'. congr.
rewrite <- H. rewrite H1. rewrite bij_left. reflexivity.
intros. subst e0.
rewrite <- H0. do 2 rewrite fct_fct. congr.
extens. rewrite H1. rewrite bij_left. reflexivity.

intros.
rename X into Z'. rename X0 into X.
rename Y into Z0. rename Y0 into Y.
rename e into g'. rename e0 into e.
rename g into f2. rename f into f1.
rename d into g. rename c into g0.
rename h into f.
apply appendenv_WF_2_xfer_trs with
  (b' := b') (b := b) (f1 := f1) (f2 := f2) (f := f)
  (g := g) (g' := g') (g0 := g0).
apply H. apply H2. apply H3. apply H4.
apply IHWF0_1. apply IHWF0_2.
Qed.

Let appendenv_WF_2_rmk : forall (X' Y' X Y : Set)
    (fx : X' -> X) (fy : Y' -> Y)
    (f : X' + Y' -> X + Y)
    (g' : env0 (X' + Y')) (g : env0 (X + Y))
    (e : env X Y) (e' : env X' Y'),
  (f = sumfct fx fy) ->
  (e' = fun y => (g' (inr X' y))) ->
  (e = fun y => g (inr X y)) ->
  WF0 f g' g ->
  WF fx fy e' e.
Proof.
intros.
apply appendenv_WF_2_xfer with
  (f := f) (g' := g') (g := g)
  (b' := id_bij (X' + Y'))
  (b := id_bij (X + Y)).
trivial.
intros. rewrite H. reflexivity.
unfold env. extens. rewrite H1. simpl. symmetry.
apply fct_id. reflexivity.

unfold env. extens. rewrite H0. simpl. symmetry.
apply fct_id. reflexivity.
Qed.

Lemma appendenv_WF_2 : forall (X' Y' X Y : Set)
    (fx : X' -> X) (fy : Y' -> Y) (f : X' + Y' -> X + Y)
    (g' : env0 X') (d' : env X' Y')
    (g : env0 X) (d : env X Y),
  (f = sumfct fx fy) ->
  WF0 f (appendenv g' d') (appendenv g d) -> WF fx fy d' d.
Proof.
intros.
apply appendenv_WF_2_rmk with
  (f := f) (g' := appendenv g' d') (g := appendenv g d).
trivial.
unfold env. extens. reflexivity.
unfold env. extens. reflexivity.
trivial.
Qed.

Lemma appendenv_WF0all_2 : forall (X Y : Set)
    (g : env0 X) (d : env X Y),
  WF0all (appendenv g d) ->
  WF (empty_inc X) (empty_inc Y) (empty_inc _) d.
Proof.
unfold WF0all. intros.
apply appendenv_WF_2 with
  (f := sumfct (empty_inc X) (empty_inc Y))
  (g' := empty_inc (T empty)) (g := g).
reflexivity.
eapply WF0_trs with
  (c := empty_inc (T (empty)))
  (g := empty_inc (X + Y))
  (f := fun u : empty + empty =>
          match u with
          | inl x => x
          | inr y => y
          end).
destruct u as [x|x]; destruct x.
apply WF0_intro with (g := empty_inc (T (empty + empty))).
destruct x as [x|x]; destruct x.
destruct y.
assumption.
Qed.

End Appendenv_WF0.

(* --------------------------------------------------------------- *)
(** * The [WFF] relation.                                          *)
(* --------------------------------------------------------------- *)

(* --------------------------------------------------------------- *)
(** Another auxiliary definition of well-formed environment.
    Produces an induction principle similar to WF but with
    "fixed X".                                                     *)
(* --------------------------------------------------------------- *)

Inductive WFF : forall X Y1 Y2 : Set,
    (Y1 -> Y2) -> env X Y1 -> env X Y2 -> Prop :=
  | WFF_intro : forall (X Y1 Y2 : Set)
      (e : env X Y1) (d : env X Y2)
      (f : Y1 -> Y2) (g : Y2 -> T (X + Y1)),
    (forall x, g (f x) = e x) ->
    (forall y, fct (sumfct_r f) (g y) = d y) -> WFF f e d
  | WFF_trs : forall X Y1 Y2 Y3
       (e : env X Y1) (c : env X Y2) (d : env X Y3)
       (f : Y1 -> Y2) (g : Y2 -> Y3) (h : Y1 -> Y3),
     (forall u, h u = g (f u)) ->
     WFF f e c -> WFF g c d -> WFF h e d.
Hint Constructors WFF.

(* --------------------------------------------------------------- *)
(** We prove that WF and WFF are equivalent (when it makes sense).
    The following lemma is the easy part of this equivalence.      *)
(* --------------------------------------------------------------- *)

Lemma WFF_WF : forall (X Y1 Y2 : Set) (f : Y1 -> Y2)
    (g : env X Y1) (d : env X Y2),
  WFF f g d -> WF (fun x : X => x) f g d.
Proof.
induction 1.
apply WF_intro with g; trivial.
pose (idX := fun x : X => x).
apply WF_trs with
    (c := c) (f := f) (g := g) (r' := idX) (r'' := idX);
  trivial.
Qed.

(* --------------------------------------------------------------- *)
(** Now the hard part of the equivalence WF WFF.                   *)
(* --------------------------------------------------------------- *)

Section WF_WFF.

Remark WFF_fct_X : forall (X Y' Y : Set)
    (fy : Y' -> Y) (g' : env X Y') (g : env X Y),
  WFF fy g' g ->
  forall (X0 : Set) (gx : X -> X0) (e' : env X0 Y') (e : env X0 Y),
  (forall y : Y', e' y = fct (sumfct_l (Y := Y') gx) (g' y)) ->
  (forall y : Y, e y = fct (sumfct_l (Y := Y) gx) (g y)) ->
  WFF fy e' e. 
Proof.
induction 1.
intros.
apply WFF_intro with
  (g := fun y : Y2 => fct (sumfct_l (Y := Y1) gx) (g y)).
intros. rewrite H1. rewrite H. reflexivity.
intros. rewrite fct_fct. rewrite H2. rewrite <- H0. rewrite fct_fct.
congr. extens x. destruct x; simpl; trivial.
intros.
apply WFF_trs with
  (c := fun y : Y2 =>
    fct (sumfct_l (Y := Y2) gx) (c y))
  (f := f) (g := g).
assumption.
apply IHWFF1 with (gx := gx).
auto.
auto.
apply IHWFF2 with (gx := gx); auto.
Qed.

Remark WF_WFF_rmk : forall (X' Y' X Y : Set)
    (fx : X' -> X) (fy : Y' -> Y)
    (g' : env X' Y') (g : env X Y) (g0 : env X Y'),
  (forall y : Y' , g0 y = fct (sumfct_l (Y := Y') fx) (g' y)) ->
  WF fx fy g' g -> WFF fy g0 g.
Proof.
induction 2.
apply WFF_intro with
  (g := fun y => fct (sumfct_l (Y := Y1) r) (g y)).
intros. rewrite H. auto.
intros.
rewrite fct_fct.
rewrite <- H1.
congr. extens x; destruct x; simpl; auto.
eapply WFF_trs with
   (c := fun y : Y2 => fct (sumfct_l (Y := Y2) r'') (c y)).
apply H0.
2: apply IHWF2; auto.
clear IHWF2.

eapply WFF_fct_X with
  (g' := fun y => fct (sumfct_l (Y:=Y1) r') (e y))
  (g := c).
apply IHWF1.
reflexivity.
intros. rewrite H. rewrite fct_fct.
congr. extens x. destruct x; simpl; auto. reflexivity.
Qed.

Lemma WF_WFF : forall (X Y1 Y2 : Set) (f : Y1 -> Y2)
    (g : env X Y1) (d : env X Y2),
  WF (fun x : X => x) f g d -> WFF f g d.
Proof.
intros.
eapply WF_WFF_rmk with (fx := fun x : X => x).
2: apply H.
intros.
symmetry.
apply fct_id.
destruct u; simpl; auto.
Qed.

End WF_WFF.

Lemma WFF_WF0_appendenv : forall (X Y1 Y2 : Set) (f : Y1 -> Y2)
    (g : env0 X) (d1 : env X Y1) (d2 : env X Y2),
  WFF f d1 d2 ->
  WF0 (sumfct_r f) (appendenv g d1) (appendenv g d2).
Proof.
induction 1.
apply WF0_intro with
  (g := fun u : X + Y2 =>
          match u with
          | inl x => fct (@inl X Y1) (g x)
          | inr y => g0 y
          end).
destruct x; simpl; trivial.
destruct y; simpl.
rewrite fct_fct. simpl. congr. extens. reflexivity.
trivial.
apply WF0_trs with
  (c := appendenv g c)
  (f := sumfct_r (X := X) f)
  (g := sumfct_r (X := X) g0).
destruct u; simpl; auto.
auto.
auto.
Qed.

Lemma WF0_appendenv_Y : forall (X Y1 Y2 : Set) (f : Y1 -> Y2)
    (g : env0 X) (d1 : env X Y1) (d2 : env X Y2),
  WF (fun x => x) f d1 d2 ->
  WF0 (sumfct_r f) (appendenv g d1) (appendenv g d2).
Proof.
intros.
apply WFF_WF0_appendenv.
apply WF_WFF.
assumption.
Qed.

(* --------------------------------------------------------------- *)
(** * Proof of WF0all_appendenv.                                   *)
(* --------------------------------------------------------------- *)

Section WF_sub_lemma.

Let WF_fct_X : forall (X' Y' X Y : Set)
    (fx : X' -> X) (fy : Y' -> Y) (g' : env X' Y') (g : env X Y),
  WF fx fy g' g ->
  forall (X0 : Set) (gx : X -> X0) (gx0 : X' -> X0)
    (e' : env X0 Y') (e : env X0 Y),
  (forall x : X', gx0 x = gx (fx x)) ->
  (forall y : Y', e' y = fct (sumfct_l (Y := Y') gx0) (g' y)) ->
  (forall y : Y, e y = fct (sumfct_l (Y := Y) gx) (g y)) ->
  WF (fun x : X0 => x) fy e' e. 
Proof.
induction 1.
intros.
apply WF_intro with
  (g := fun y : Y2 => fct (sumfct_l (Y := Y1) gx0) (g y)).
intros. rewrite H2. rewrite H. reflexivity.
intros. rewrite fct_fct. rewrite H3. rewrite <- H0. rewrite fct_fct.
congr. extens x. destruct x; simpl. rewrite H1. reflexivity. reflexivity.
intros.
apply WF_trs with
  (c := fun y : Y2 =>
    fct (sumfct_l (Y := Y2) (fun x => gx (r'' x))) (c y))
  (r' := fun x : X0 => x)
  (r'' := fun x : X0 => x)
  (f := f) (g := g).
assumption.
reflexivity.
apply IHWF1 with (gx := fun x => gx (r'' x)) (gx0 := fun x => gx (r x)).
auto.
intros. rewrite H4.
congr. extens x. destruct x; simpl; auto.
reflexivity.
apply IHWF2 with (gx := gx) (gx0 := fun x => gx (r'' x)).
reflexivity.
reflexivity.
intros.
rewrite H5.
reflexivity.
Qed.

Lemma WF_sub_lemma : forall (X' Y' X Y : Set)
    (fx : X' -> X) (fy : Y' -> Y)
    (d' : env X' Y') (d : env X Y),
  WF fx fy d' d ->
  forall (d0 : env X Y'),
  (forall y, d0 y = fct (sumfct_l (Y := Y') fx) (d' y)) ->
  WF (fun x : X => x) fy d0 d.
Proof.
induction 1.
intros.
apply WF_intro with
  (g := fun y : Y2 => fct (sumfct_l (Y := Y1) r) (g y)).
intros. rewrite H1. rewrite H. reflexivity.
intros. rewrite fct_fct. rewrite <- H0.
congr. extens x. destruct x; reflexivity.
intros.
eapply WF_trs with
  (c := fun y => fct (sumfct_l (Y := Y2) r'') (c y))
  (r'' := fun x : X3 => x)
  (r' := fun x : X3 => x)
  (f := f) (g := g).
apply H.
reflexivity.

2: apply IHWF2; reflexivity.

clear IHWF2.
eapply WF_fct_X with (gx := r'') (gx0 := r) (fx := r').
apply H1.
auto.
auto.
reflexivity.
Qed.

End WF_sub_lemma.

Section WF0_appendenv.

Let WF0_appendenv_X : forall (X' Y' X: Set)
    (fx : X' -> X)
    (g' : env0 X') (g : env0 X),
  WF0 fx g' g ->
    forall (f : X' + Y' -> X + Y')
      (d' : env X' Y') (d : env X Y'),
    (f = sumfct_l fx) ->
    (forall y : Y', d y = fct (sumfct_l fx) (d' y)) ->
    WF0 f (appendenv g' d') (appendenv g d).
Proof.
induction 1; intros.
apply WF0_intro with
  (g := fun u =>
          match u with
          | inl x => fct (inl Y') (g x)
          | inr y => d' y
          end).
rewrite H1. destruct x; simpl; auto.
destruct y; simpl.
rewrite <- H0. do 2 rewrite fct_fct.
congr. extens. rewrite H1. reflexivity.
rewrite H2. auto.
rename f into f1. rename g into f2.
rename X into X'. rename Y into X0. rename Z into X.
rename e into g'. rename d into g. rename d0 into d.
pose (d0 := fun y : Y' => fct (sumfct_l f1) (d' y)).
apply WF0_trs with
  (c := appendenv c d0)
  (f := sumfct_l (Y := Y') f1)
  (g := sumfct_l (Y := Y') f2).
intros.
rewrite H2.
destruct u; simpl. rewrite H. reflexivity. reflexivity.
apply IHWF0_1; auto.
apply IHWF0_2.
reflexivity.
intros. rewrite H3. unfold d0. rewrite fct_fct.
congr. extens x. destruct x; simpl; auto.
Qed.

Lemma WF0_appendenv : forall (X' Y' X Y : Set)
    (fx : X' -> X) (fy : Y' -> Y) (f : X' + Y' -> X + Y)
    (g' : env0 X') (g : env0 X) (e' : env X' Y') (e : env X Y),
  (f = sumfct fx fy) ->
  WF0 fx g' g -> WF fx fy e' e ->
  WF0 f (appendenv g' e') (appendenv g e).
Proof.
intros.
apply WF0_trs with
  (c := appendenv g (fun y : Y' => fct (sumfct_l fx) (e' y)))
  (f := sumfct_l (Y := Y') fx)
  (g := sumfct_r (X := X) fy).
intros. rewrite H.
destruct u; reflexivity.
apply WF0_appendenv_X with (fx := fx).
assumption.
reflexivity.
reflexivity.
apply WF0_appendenv_Y.
eapply WF_sub_lemma.
apply H1.
reflexivity.
Qed.

End WF0_appendenv.

Lemma WF0all_appendenv : forall (X Y : Set) (g : env0 X) (d : env X Y),
  WF0all g ->
  WF (empty_inc _) (empty_inc Y) (empty_inc (T (empty + empty))) d ->
  WF0all (appendenv g d).
Proof.
intros.
unfold WF0all.
apply WF0_trs with
  (c := appendenv (empty_inc (T empty)) (empty_inc (T (empty + empty))))
  (f := empty_inc (empty + empty))
  (g := sumfct (empty_inc X) (empty_inc Y)).
destruct u.
eapply WF0_intro
  with (g := fun _ : empty + empty => top empty).
destruct x.
destruct y as [e|e]; destruct e.
apply WF0_appendenv with (fx := empty_inc X) (fy := empty_inc Y).
reflexivity.
apply H.
apply H0.
Qed.
